/*
*   Projection
*
*   Class which adapts a projected image to the topology of the sandbox and the beamer point of view.
*   Contains the matrix to compensate for the rotation of the beamer
*   and the distance to the plane at the top of the sandbox.
*/

#include "../../inc/projection.h"

/*
 *   MAIN
 */

Projection::Projection(){
    adjustingMatrix = cv::getRotationMatrix2D(cv::Point(0, 0), 0, 1);
    distanceTopSandbox = 1.0f;
}


/*
*   Adjust the projected frame with the topology from the camera to the beamer POV (point of view)
*
*   depth       : Topology of the sandbox under the projection
*   src         : Image source projected to adjust to the topology
*   dst         : Output which will contain the adapted image
*   camera      : Active camera
*   beamer_pos  : 3D position of the beamer relative to the camera
*/
void Projection::adjustFrame(cv::Mat_<float> &depth, cv::Mat_<cv::Vec3b> &src, cv::Mat_<cv::Vec3b> &dst, Camera *camera, cv::Point3f beamer_pos){
    
    // update if dst.size changed, since for all buffers :
    //      buff.size == dst.size
    if(deprojectMap.size() != dst.size()){
        std::vector<cv::Point2f> profil = camera->getAdaptedIntrinsics(dst);
        fxy = profil.at(0);
        ppxy = profil.at(1);
    }
    // do nothing if contains matrix with same size
    // otherwise => release => allocate
    deprojectMap.create(dst.rows, dst.cols);
    frameMap.create(dst.rows, dst.cols);
    resized_depth.create(dst.rows, dst.cols);
    resized_src.create(dst.rows, dst.cols);

    deprojectMap = cv::Point2i(-1,-1);
    frameMap = cv::Point2i(-1,-1);

    // resize to match 1:1 ratio with dst, since we'll do later:
    //      dst[i] = src[i]
    cv::resize(src, resized_src, dst.size());
    cv::resize(depth, resized_depth, dst.size());    

    deprojectPixelsFromDepth(resized_depth, camera, camera->getCroppingMask() , beamer_pos, deprojectMap, fxy, ppxy);
    filterLowestDeprojectedPoints(resized_depth, deprojectMap, frameMap);
    buildFrame(resized_depth, frameMap, resized_src, dst);
    holeFilling(dst, frameMap);
    cv::warpAffine(dst, dst, adjustingMatrix, dst.size()); // apply the rotation on the image
}



/*
 *   Private
 */


/*
*   Deproject pixels in 3D, then adapt to Beamer's POV, and go back to 2D
*   This gives us the location of pixels adapted to the Beamer projection
*
*   depth           : Topology of the sandbox under the projection
*   camera          : Active camera
*   mask            : ROI (Range Of Interest) delimiting the zone of the projection from the camera POV
*   beamer_pos      : 3D position of the beamer relative to the camera
*   deprojectMap    : Indicates for each pixel of src, where it'll be displayed
*   fxy             : function x and y adapted to the projection matching the original depth matrix(without ROI) of the camera
*   ppxy            : coordinates x and y of the central pixel adapted to the projection matching the original depth matrix(without ROI) of the camera
*/
void Projection::deprojectPixelsFromDepth(cv::Mat_<float> &depth, Camera *camera, cv::Rect mask, cv::Point3f beamer_pos, cv::Mat_<cv::Point2i> &deprojectMap, cv::Point2f fxy, cv::Point2f ppxy){
    
    // scale coord of the mask with the new resolution of the depth frame
    float mask_scale_x = mask.x * depth.cols / mask.width;
    float mask_scale_y = mask.y * depth.rows / mask.height;

    for (int i = 0; i < depth.rows; i++){
        for (int j = 0; j < depth.cols; j++){
            // coordinates of the pixel relative to the orginial image taken from the camera
            int base_x = mask_scale_x+j;
            int base_y = mask_scale_y+i;
            float z = depth.at<float>(i,j);
            
            // pixels based on the original depth frame taken from the camera
            cv::Point2i pixel = findMatchingPixel( base_y, base_x, z, camera, beamer_pos, fxy, ppxy );
            
            // pixel relative to the cropping mask (the area where the frame is projected)
            pixel.x -= mask_scale_x;
            pixel.y -= mask_scale_y;

            deprojectMap.at<cv::Point2i>(i,j) = pixel;
        }
    }
}

/*
*   Save the highest points in deprojectMap into frameMap,
*   because some points can be deprojected at the same location
*
*   depth           : Topology of the sandbox matching the projection
*   deprojectMap    : Indicates for each pixel of src, where it'll be displayed
*   frameMap        : Indicates for each pixel of dst, where it should get the value from in src    
*/
void Projection::filterLowestDeprojectedPoints(cv::Mat_<float> &depth, cv::Mat_<cv::Point2i> &deprojectMap, cv::Mat_<cv::Point2i> &frameMap){

    for (int i = 0; i < deprojectMap.rows; i++){
        for (int j = 0; j < deprojectMap.cols; j++){

            // coords of the new pixel
            cv::Point2i deprojectedPixel = deprojectMap.at<cv::Point2i>(i,j);
            cv::Point2i highestDepthPixel = cv::Point2i(j,i);

            if( (0 <= deprojectedPixel.x && deprojectedPixel.x < depth.cols) &&
                (0 <= deprojectedPixel.y && deprojectedPixel.y < depth.rows) ){
                
                // check and keep the highest point at the location pointed by pixel
                cv::Point2i currentDepthPixel = frameMap.at<cv::Point2i>(deprojectedPixel);
                if( (0 <= currentDepthPixel.x && currentDepthPixel.x < depth.cols) &&
                    (0 <= currentDepthPixel.y && currentDepthPixel.y < depth.rows) ){
                        if(depth.at<float>(currentDepthPixel) <= depth.at<float>(i,j)){
                            highestDepthPixel = currentDepthPixel;
                        }
                }
                frameMap.at<cv::Point2i>(deprojectedPixel) = highestDepthPixel;
            }
        }
    }
}


/*
*   Build the frame using frameMap, we assume that all the buffers have the same size
*   where each pixel describes in which pixel of the source it should take the value from
*       dst[i] = src[frameMap[i]]
*
*   frameMap    : The map describing where are the source pixels for each pixel in our output image
*   src         : Image source to adapt
*   dst         : Output image adapted from src to project
*/
// TODO : enlever depth en paramètre et vérifier que le pixel soit dans la range d'un des autres buffer
void Projection::buildFrame(cv::Mat_<float> &depth, cv::Mat_<cv::Point2i> &frameMap, cv::Mat_<cv::Vec3b> &src, cv::Mat_<cv::Vec3b> &dst){
    
    for (int i = 0; i < frameMap.rows; i++){
        for (int j = 0; j < frameMap.cols; j++){

            cv::Point2i pixel_src = frameMap.at<cv::Point2i>(i,j);
            cv::Point2i pixel_dst = cv::Point2i(j,i);

            if( (0<=pixel_src.x && pixel_src.x<depth.cols) && (0<=pixel_src.y && pixel_src.y<depth.rows) ){
                // src and dst must be of same size
                dst.at<cv::Vec3b>(pixel_dst) = src.at<cv::Vec3b>(pixel_src);
            }
        }
    }
}


/*
*   Fixe holes formed by the deprojection due to round up coordinates
*   (because deproject goes from 3D floats values to 2D uint),
*   by filling with value of the 1st non null neighbour
*
*   dst         : The output image to project
*   frameMap    : The map describing where are the source pixels for each pixel in our output image
*/
void Projection::holeFilling(cv::Mat_<cv::Vec3b> &dst, cv::Mat_<cv::Point2i> &frameMap){

    for (int i = 0; i < dst.rows; i++){
        for (int j = 0; j < dst.cols; j++){
            
            if(frameMap.at<cv::Point2i>(i,j) == cv::Point2i(-1,-1)){

                cv::Vec3b color;
                bool found = false;

                for(int d_y = -1; d_y <= 1; d_y++){
                    for(int d_x = -1; d_x <= 1; d_x++){
                
                        if(!(d_x == 0 && d_y == 0)){
                            
                            int y = i+d_y;
                            int x = j+d_x;
                            if( ((0 <= x && x < dst.cols) && (0 <= y && y < dst.rows)) && (frameMap.at<cv::Point2i>(y,x) != cv::Point2i(-1,-1)) ){
                                color = dst.at<cv::Vec3b>(y,x);
                                found = true;
                                break;
                            }
                        }
                    }
                    if(found)
                        break;
                }

                if(found){
                    dst.at<cv::Vec3b>(i,j) = color;
                }
            }
        }
    }
}


/*
*    C : Camera position
*    B : Beamer position
*    P : Point computed by camera depth
*    V : Point at the plane height adjusted to the beamer POV (Point Of View)
*    E : Point of the right-angle triangle VEB at the plane height 
*    A : Point of the right-angle triangle PAB at the point P height
*    
*    Where
*        CP : distance from camera to point (value of depth_frame)
*        CB : distance from camera to beamer (beamer's position is relative to the camera)
*
*   i       : y coordinate of the pixel to adjust
*   j       : x coordinate of the pixel to adjust
*   z       : Z coordinate of the pixel from the depth matrix
*   camera  : Camera active
*   CB      : Position of the beamer relative to the camera (vector camera-beamer)
*   fxy     : function x and y adapted to the projection matching the original depth matrix(without ROI) of the camera
*   ppxy    : coordinates x and y of the central pixel adapted to the projection matching the original depth matrix(without ROI) of the camera
*
*   Return the coordinates of the pixel source adapted to the beamer POV
*/
cv::Point2i Projection::findMatchingPixel(int i, int j, float z, Camera *camera, cv::Point3f CB, cv::Point2f fxy, cv::Point2f ppxy){
   
    float pixel[2] = {static_cast<float>(j), static_cast<float>(i)};

    cv::Point3f CP = camera->deprojectPixelToPoint(pixel, z, fxy, ppxy);
    cv::Point3f BP = CP - CB;

    const float BEz = distanceTopSandbox - CB.z;
    float BAz = BP.z;
    float alpha = BEz / BAz;

    cv::Point3f BV = (alpha * BP);
    cv::Point3f CV = CB + BV;

    return camera->projectPointToPixel(CV, fxy, ppxy);
}






/*
*   Sandbox Setup purpose
*/

// TODO : move rotatePixel and revertRotatePixel in SandboxSetup (they don't depend on Projection anymore)


/*
*   Rotate a pixel to compensate for the rotate of the beamer
*/
cv::Point2i Projection::rotatePixel(cv::Point2i center, double angle, cv::Point2i pixel){

    cv::Mat_<float> matRotation = cv::getRotationMatrix2D(center, angle, 1);
    cv::Mat tmp = (cv::Mat_<cv::Vec2f>(1, 1) << cv::Vec2f(pixel.x, pixel.y));
    cv::transform(tmp, tmp, matRotation);
    return cv::Point2i(tmp.at<cv::Vec2f>(0, 0));
}

/*
*   Rotate back a rotated pixel to match the projection of the beamer
*/
cv::Point2i Projection::revertRotatePixel(cv::Point2i center, double angle, cv::Point2i rotatedPixel){

    cv::Mat_<float> matRotation = cv::getRotationMatrix2D(center, angle, 1);
    cv::Mat tmp = (cv::Mat_<cv::Vec2f>(1, 1) << cv::Vec2f(rotatedPixel.x, rotatedPixel.y));
    cv::Mat invMat;
    cv::invertAffineTransform(matRotation, invMat);
    cv::transform(tmp, tmp, invMat);
    return cv::Point2i(tmp.at<cv::Vec2f>(0, 0));
}






/*
*   Debug
*/

void Projection::printAdjustingMatrix(){

    cv::Mat matrix = getAdjustingMatrix();
    for (int y = 0; y < matrix.rows; y++){
        for (int x = 0; x < matrix.cols; x++){
            std::cout << (float)matrix.at<double>(y, x) << " ";
        }
        std::cout << std::endl;
    }
}